/* SC Dense Code. -- 
A semistatic word-based byte oriented compressor for text files.

Programmed by Antonio Faria
Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*-----------------------------------------------------------------------
 File       : Compressor of SCDC. main file.
 ------------------------------------------------------------------------*/

#include "compressor.h"


/*------------------------------------------------------------------
 Function used by qsort to compare two elements a and b
------------------------------------------------------------------*/
int comparaFrecListaDesc(const void *a, const void *b)
{
	unsigned long *left,*right;
	left =  (unsigned long *) a;
	right = (unsigned long *) b;
    if ( hash[*left].weight < hash[*right].weight)
                return 1;
    if ( hash[*left].weight > hash[*right].weight)
                return -1;

    return 0;
}

/*------------------------------------------------------------------
 Initilizes the structures used.
 ---------------------------------------------------------------- */
void initialize (unsigned long sizeVoc){
	unsigned long i;

	_memMgr = createMemoryManager();
    initialize_hash(sizeVoc);

    positionInTH = (unsigned long*) malloc (sizeVoc * sizeof(unsigned long));
	zeroNode = 0;
}

/*------------------------------------------------------------------
 Generates sequentially all the codes of the sorted vocabulary.
 ---------------------------------------------------------------- */
void GeneraCodigosSCNumero (unsigned long int v, int s)
{
	register int c;
	register unsigned long i;

	register unsigned long x,y,z,t;	// codeword bytes -> [x][y][z][t]
	unsigned long p0,p1,p2,p3;
	unsigned long base[4];

	unsigned long n;    		//codeword as a number
	register int k;				//codeword lenght.

	p0=256*256*256;
	p1=256*256;
	p2=256;
	p3=-1;   //no se usa

	c=MAX-s;

	//const unsigned long TABLABASE[5] = {0,128,16512,2113664,270549120};

	base[3]=s;
	base[2]=s*c + base[3];
	base[1]= s*c*c + base[2];
	base[0]=-1;

	i=0;
	k=1;

 	x=0; y=0; z=0;t=0;

	//x=s;
	while ( (i<v) && (x<MAX) ) {
		//y=s;
		while ( (i<v) && (y<MAX) ) {
			//z=s;
			while ( (i<v) && (z<MAX) ) {
				//t=0;
				n=x*p0 + y*p1 + z*p2 + t;
				while ( (i<v) && (t<s) ) {
					hash[positionInTH[i]].codeword=n;
					hash[positionInTH[i]].size=k;
					//fprintf(stderr,"\n word [%s], codeword [%d]",ht.hash[positionInTH[i]].word,ht.hash[positionInTH[i]].codeword);

					n++;
					i++;
					t++;
				}
				if (i!=base[3]) {z++;} else{z=s;k++;}
				t=0;
				//z++;
			}
			if ( i!= base[2] ) {y++;} else {y=s;z=s;k++;}
			z=s;
			//y++;
		}
		if (i != base[1]) {x++;} else {x=s;y=s;z=s;k++;}
		y=s;
		//x++;
	}
}


//	void GeneraCodigosSCOLD (t_NodohashTable *listaPalabras,unsigned long int v, int s)
//	
//	{
//	register int c;
//	register int MINT;			//primer valor de Terminal  (256-s)
//	unsigned long minKBytes;	//posicin de la 1 que se codifica con K bytes
//	unsigned long totalKBytes;	//total cdigos que se codifican con K bytes.
//	register int k;				//longitud de cdigo en cada momento.
//	register int i;
//	register unsigned long p,paux;	//indices de bucle (palabra sobre la que se trabaja).
//	//MAX = 256
//	
//		MINT = MAX-s;
//		c=MAX-s;
//		minKBytes=0;
//		totalKBytes=s;
//	
//		p=0;
//		k=1;
//	
//		while (p<v) {
//			paux=p-minKBytes;
//			while ((p<v) && (paux < totalKBytes)) {
//	
//				listaPalabras[p].codigo[k]  = '\0';
//				listaPalabras[p].codigo[k-1]= ( unsigned char) ((paux % s) + MINT );
//	
//				paux=paux /s;
//	
//				for (i= k-2 ; i>=0; i--) {
//					listaPalabras[p].codigo[i] = (unsigned char) (paux %c);
//					paux = paux / c;
//				}
//	
//				(listaPalabras)[p].numBytesCodigo = k;
//	
//				//Tabla = insertar_Hash(&(listaPalabras[p]), tam_tabla, Tabla);
//				p++;
//				paux = p - minKBytes;
//			}
//	
//			k++;
//			minKBytes = minKBytes + totalKBytes;
//			totalKBytes = totalKBytes * c;
//		}
//	}
//	
//	

/*------------------------------------------------------------------
Given s, c, and a list of accumulated frequencies, obtains the size of the compressed file (in bytes)
 ---------------------------------------------------------------- */
unsigned long int computeSizeS(int s, int c, long *freqList, unsigned long n, unsigned long SIZEBEST)
{
	unsigned long int totFreq,potencia,Left,total;
	unsigned long int i;

    totFreq = freqList[n];
	total = 0;
    potencia =s;
	Left = 0;
			//      i=0;   -- DEBUGGING
	while ((Left <n) && (total < SIZEBEST)) {
		total += totFreq - freqList[Left];
		Left  += potencia;
		potencia *= c;
			//      i++;
			/*              if (total>=SIZEBEST){
						//      printf("breaking K=%d ",i);
								break;
						}
			//              fprintf(stderr,"s=%d K=%d ",s,i);
			*/
	}
	return total;
}

/*------------------------------------------------------------------
Returns the optimal value of S for a given vocabulary.
 ---------------------------------------------------------------- */
int findBestS_Secuencial(int B, long int *freqList, unsigned long int longFrec) {
	unsigned long i, BestS;
	unsigned long sizeBest, sizeI;

	i = 1;
	sizeBest = computeSizeS(i, B-i, freqList, longFrec,MAX_UNSIGNED_LONG_INT);
	BestS=1;

	for (i=2; i<MAX-1; i++) {
		sizeI = computeSizeS(i, B-i, freqList, longFrec,sizeBest);
		if (sizeI < sizeBest) {
			sizeBest=sizeI;
			BestS=i;
		}
	}

	return BestS;
}

/*------------------------------------------------------------------
 * Performs the full encoding process.
   Two passes through the text:
		1- first pass: gather words and frequencies.
		  * Computing the optimal (s,c) values.
		  * Encoding process performed in the vocabulary
		2- second pass:
		  * Replaces words by codewords.
	Outputs:
		- compressed file
		- file with the vocabulary.
------------------------------------------------------------------ */
void mainCompressor (char* infile, char* outfile, unsigned long N_val) {

	unsigned long codeword,tam;
	BufferRead B;
	unsigned char *text,*top, *nextbuf;
	register unsigned char *pbeg,*pend,*wordstart;
	register unsigned long size;
	BufferWrite bw,voca;

	register unsigned long k,i,j;
	unsigned long *listaFrecuencias;

	int f;
	char *voc;
	unsigned char *aWord;
	int optimalS;
	unsigned char valor_S[4];

    voc = (char *) malloc(strlen(outfile)+256); //name of the vocabulary file

    StartValid();

	// Initialization of read buffer
	f = open (infile, O_RDONLY);
	if (f == -1) {
		printf("Cannot read file %s",infile);
		exit(0);
	}

	B = bufCreate(MAX_SIZE_OF_WORD+5);//buffer will have at least 5 bytes after a word

	bufSetFile (B,f);

	initialize(N_val); //initializes compressor for up to N_value words

	fprintf(stderr, "\nSTARTING COMPRESSION...");

	/********************** Beginning of first pass **********************/

	if (bufEmpty(B) ) {size =0; return;}

	size = 0;

	while ( B->size==B->dsize ) {

		bufCurrent (B,(char**)&text,(char**)&top);
		pbeg = text; pend = top;

	 	while ( pbeg < pend )   {

			//parsing a word or separator
			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data

				while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] )) {
					size++;
					pbeg++;
				}

		    }
			else {
				if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word
					//fprintf(stderr,"\n Entra 3 pbeg [%d] pend [%d] valid [%d]",pbeg,pend, _Valid[*pbeg]);
					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])) {

						size++;
						pbeg++;
					}
				}
				else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not

	                pbeg++;
					if (_Valid [*pbeg] ) {
						wordstart = pbeg;
						while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}

					}
					else {   // a "separator word" is parsed until an alphanumerical character is found

						size++;
						while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}
					}//else {  // a "separator word"
				}//if (!bufEmpty(B)) {
			}

			aWord=wordstart;  //the word parsed.

	        j = search ((unsigned char *)aWord, size, &addrInTH );

	        if (j==zeroNode) {
				insertElement ((unsigned char *) aWord, size, &addrInTH);
				hash[addrInTH].weight = 0;
				hash[addrInTH].size = 0;
				hash[addrInTH].len=size;
				positionInTH[zeroNode] = addrInTH;
				zeroNode++;
			}

			hash[addrInTH].weight +=1;

    	}/* internal while*/

		nextbuf=pbeg;
		bufLoad (B, (char*)nextbuf);
	}

	//processing the last buffer.

	if (!bufEmpty(B)){
	 	bufCurrent (B,(char**)&text,(char**)&top);
	 	pbeg = text; pend = top;

		while ( pbeg < pend )   {

			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
		    	while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] ) &&( pbeg < pend )) {

						size++;
						pbeg++;
				}
		    }
			else {
				if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word

					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])&&( pbeg < pend )) {

						size++;
						pbeg++;
					}
				}
				else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not

				//if (!bufEof(B)) { //not end of file
	                pbeg++;
					if (_Valid [*pbeg] ) {
						wordstart = pbeg;
						while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )&&( pbeg < pend )) {
							size++;
							pbeg++;
						}

					}
					else {   // a "separator word" is parsed until an alphanumerical character is found

						size++;
						while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )&&( pbeg < pend )) {
							size++;
							pbeg++;
						}
					}//else {  // a "separator word"
				}//if (!bufEmpty(B)) {

			}

	  		if ( pbeg >= pend ) {  //needed in case of the last word been parsed further away of cend. (up to MAX_SIZE_OF_WORD positions)
	  			size -= (pbeg-pend);	
	  		}

			aWord=wordstart;

			j = search ((unsigned char *)aWord, size, &addrInTH );

			if (j==zeroNode) {
				insertElement ((unsigned char *) aWord, size, &addrInTH);
				hash[addrInTH].weight = 0;
				hash[addrInTH].size = 0;
				hash[addrInTH].len=size;
				positionInTH[zeroNode] = addrInTH;
				zeroNode++;
		   	}

			hash[addrInTH].weight +=1;

    	}/* while interno*/

	}//if (!bufEmpty(B){


	fprintf(stderr,"\n Number of words in the vocabulary:  %ld \n",zeroNode);

	/********************** end of first pass **********************/

	strcpy(voc,outfile);
    strcat(voc,".voc");

	bw= bufCreateWriter(outfile);
	voca=bufCreateWriter(voc);
	openBW(bw);
	openBW(voca);

	// Sorting the vocabulary by frequency.

	{	//Moves all the words with frequency = 1 to the end of the vocabulary.
		register long ii;
		register long kk;

	    kk=zeroNode-1;
	    ii=0;
	    while (ii<kk){
			while ((hash[positionInTH[ii]].weight!=1) && (ii <kk)) { ii++; }
			while ((hash[positionInTH[kk]].weight==1) && (ii <kk)) { kk--; }

			if (ii<kk){
				swap(&positionInTH[ii], &positionInTH[kk]);
				kk--;
				ii++;
			}
	  	}
	
		//k=ii; 
		k=ii+1; //the lenght of the vector to be sorted with qsort. So v[0 .. k-1]
	}

	//Aplies qsort to the words with frequency > 1.
	qsort(positionInTH,k,sizeof(unsigned long),comparaFrecListaDesc);

	//Computes the optimal (s,c) values.
    listaFrecuencias = (unsigned long *) malloc ((zeroNode+1)*sizeof(unsigned long));

   // List of accumulated frecuencies used by findBestS.
    listaFrecuencias[0] = 0;
	for (i=0;i<zeroNode;i++){
			listaFrecuencias[i+1] =listaFrecuencias[i] + hash[positionInTH[i]].weight;
	}

   optimalS = findBestS_Secuencial(MAX, listaFrecuencias, zeroNode);
	//Generates codes sequentially

	GeneraCodigosSCNumero(zeroNode,optimalS);

	//for (i=0;i<zeroNode;i++)
	// fprintf(stderr,"\n Palabra [%s], peso [%ld]",hash[positionInTH[i]].word,hash[positionInTH[i]].weight);

	//stores the number of words of the vocabulary
	writeBW(voca,(unsigned char *)&zeroNode,4);
	//stores the value of S
	sprintf(valor_S,"%i",optimalS);
	writeBW(voca,(unsigned char *)valor_S,strlen(valor_S)+1);

	//Writes the vocabulary to disk. (all the words in the sorted vocabulary)
	for (i=0;i<zeroNode;i++)
		writeBW(voca,(unsigned char *)hash[positionInTH[i]].word,hash[positionInTH[i]].len+1);

	close(f);

	f = open (infile, O_RDONLY);
	if (f == -1)
		printf("Cannot read file %s",infile);

	/********************** Beginning of the second pass **********************/

	bufSetFile (B,f);

	if (bufEmpty(B) ) {size =0; return;}
	size = 0;
	while ( B->size==B->dsize ) {

		bufCurrent (B,(char**)&text,(char**)&top);
		pbeg = text; pend = top;

	 	while ( pbeg < pend )   {

			//parsing a word or separator
			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
				while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] )) {
					size++;
					pbeg++;
				}

		    }
			else {
				if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word

					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])) {

						size++;
						pbeg++;
					}
				}
				else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not

	                pbeg++;
					if (_Valid [*pbeg] ) {
						wordstart = pbeg;
						while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}
					}
					else {   // a "separator word" is parsed until an alphanumerical character is found

						size++;
						while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}
					}//else {  // a "separator word"
				}//if (!bufEmpty(B)) {

			}

			aWord=wordstart; //the parsed word

			j = search ((unsigned char *)aWord, size, &addrInTH );

			codeword = hash[addrInTH].codeword;
			tam=hash[addrInTH].size;
			writeCodeWord(bw,&codeword,&tam);

    	}/* while interno*/
    	nextbuf=pbeg;
        bufLoad (B, (char*)nextbuf);
	}//end while


	//processing the last buffer.
	if (!bufEmpty(B)){
	 	bufCurrent (B,(char**)&text,(char**)&top);
	 	pbeg = text; pend = top;

		while ( pbeg < pend )   {

			//parses a word
			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
		    	while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] ) &&( pbeg < pend )) {
					size++;
					pbeg++;
				}
		    }
			else {
				if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word
					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])&&( pbeg < pend )) {
						size++;
						pbeg++;
					}
				}
				else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not

	                pbeg++;
					if (_Valid [*pbeg] ) {
						wordstart = pbeg;
						while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )&&( pbeg < pend )) {
							size++;
							pbeg++;
						}
					}
					else {   // a "separator word" is parsed until an alphanumerical character is found

						size++;
						while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )&&( pbeg < pend )) {
							size++;
							pbeg++;
						}
					}//else {  // a "separator word"
				}//if (!bufEmpty(B)) {
			}

			aWord=wordstart;  //the parsed word

			j = search ((unsigned char *)aWord, size, &addrInTH );
			codeword = hash[addrInTH].codeword;
			tam=hash[addrInTH].size;
			writeCodeWord(bw,&codeword,&tam);
    	}/* inner while */
	}

	/********************** end of second pass **********************/

	bufDestroy(B);  //freeing resources.
	closeBW(bw);
	destroyBW(bw);
	closeBW(voca);
	destroyBW(voca);	
	free(voc);

	free(listaFrecuencias);
	free(hash);
	
	free(positionInTH);
	destroyMemoryManager(_memMgr);
} //mainDecompressor


/*------------------------------------------------------------------
 Obtains the size of the file to compress.
------------------------------------------------------------------ */
unsigned long fileSize (char *filename) {
	FILE *fpText;
	unsigned long fsize;
	fpText = fopen(filename,"rb");
	fsize=0;
	if (fpText) {
		fseek(fpText,0,2);
		fsize= ftell(fpText);
		fclose(fpText);
		fprintf(stderr,"fileSize = %ld",fsize);
	}
	return fsize;
}


/* Shows a message with info of this version of the code */
void showMessage() {
	printf("\n*(S,C)-Dense Code v0.9 [semistatic compressor]");
	printf("\n*CopyRight(c)2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param\n");
	
    printf("\n  #SCDC comes with ABSOLUTELY NO WARRANTY; ");
    printf("\n  #This is free software, and you are welcome to redistribute it ");
    printf("\n  #under certain conditions\n\n");
}

/*------------------------------------------------------------------ */
/*
 *  COMPRESSOR PROGRAM OF SCDC.
 */
/*------------------------------------------------------------------ */

int main(int argc, char* argv[])
{
	char fileName[20];
	long sizeNValue;
	unsigned long bytesFile;

	showMessage();


	//printf("ARGC VALE [%d]\n",argc);
	if (argc < 3) {
		fprintf(stderr,"\n Wrong call to SCDC compressor\n");
		fprintf(stderr," Use:  %s  <in file>  <out file> <[VOC SIZE]> \n",argv[0]);
		return 0;
	}

	//Estimation (Using Heap's law) of the size of the vocabulary (in number of distinct words).
	//sizeNValue=N_value;
	bytesFile = fileSize(argv[1]);
	if (!bytesFile) {
		fprintf(stderr, "\nFILE EMPTY OR FILE NOT FOUND !!\nCompression aborted\n");
		exit(0);
	}

	if(bytesFile<5000000) bytesFile = 5000000;
	sizeNValue = (unsigned long) floor(3.9* pow(bytesFile,0.69) );

	if (argc ==4) {
		unsigned long sizetmp;
		sizetmp = atol(argv[3]);
		//if (sizetmp > sizeNValue) {sizeNValue = sizetmp;}
		sizeNValue = sizetmp;
	}

	fprintf(stderr,"\nMax size of voc allowed %ld",sizeNValue);

	mainCompressor(argv[1], argv[2],sizeNValue);
	StopRusageTime("timesComp",argv[1]);
	fprintf(stderr,"\nExact Number of words in vocabulary: %ld ",zeroNode);
	fprintf(stderr,"\n\n*Compression finished.\n");
}
