/* SC Dense Code. -- 
A semistatic word-based byte oriented compressor for text files.

Programmed by Antonio Faria
Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*-----------------------------------------------------------------------
 File           : decompressor.c
 Function   : Main file of the decompressor of SCDC
 ------------------------------------------------------------------------*/

#include "decompressor.h"


/*------------------------------------------------------------------
 Initialization of structures used.
 ---------------------------------------------------------------- */
void initialize (unsigned long sizeVoc){
	unsigned long i;

	_memMgr = createMemoryManager(); //_memMgr.initialize();

	TABLABASE[0] = 0;
	TABLABASE[1] = S;
	TABLABASE[2] = TABLABASE[1] + S * C;
	TABLABASE[3] = TABLABASE[2] + S * C*C;
	TABLABASE[4] = TABLABASE[3] + S * C*C*C;
    //fprintf(stderr,"\n S[%ld],3[%ld],4[%ld],5[%ld]",S,TABLABASE[3],TABLABASE[4],TABLABASE[5]);

	N=sizeVoc;
	words  = (t_word *) malloc(N * sizeof(t_word));
}



/*-----------------------------------------------------------------------
 Inserts a new element into the vocabulary (in the array of words) (function not used now)
 ---------------------------------------------------------------------*/
 void insertElement (const unsigned char *aWord, register unsigned long len , register unsigned long position) {

	// if (position == N){
		// printf("Not enough memory, vocabulary exceeds maximun size !\n");
		// exit(1);
	// }

	getMemoryBlock(_memMgr,( char **)&(words[position].word),len+1);
	{
		unsigned char *w,*src;
		register int i;
		src=(unsigned char*) aWord;

		w=words[position].word;
		for (i=0;i<=len;i++) {  //the final '\0' is also copied
			*(w++)=*(src++);
		}
	}

	words[position].size=len;
}

/*------------------------------------------------------------------
 * Performs the full decoding process.
------------------------------------------------------------------ */
void mainDecompressor (char* infile, char* outfile) {

	unsigned char *palabra;
	unsigned char *tmpW;
	register unsigned char abyte;
	BufferRead B,V;
	register unsigned long j;

	int f,v;
	/**/unsigned char *text,*top, *nextbuf;
	 char *voc,valor_S[4];

	voc = (char *) malloc(strlen(infile)+256);
	strcpy(voc,infile);
    strcat(voc,".voc");


	// Inicialization of the buffers used to read the files
	f = open (infile, O_RDONLY);
	if (f == -1) {
		fprintf(stderr,"Cannot read file %s\nDecompression aborted.\n",infile);
		exit(0);
	}
	v = open (voc, O_RDONLY);
	if (v == -1) {
		fprintf(stderr,"Cannot read file %s\nDecompression aborted.\n",voc);
		exit(0);
	}

	B = bufCreate(MAX_SIZE_OF_WORD+5);    //Number of byte left at the end of a memory buffer to avoid
	V = bufCreate(MAX_SIZE_OF_WORD+5);	  //a word being "cut"

	bufSetFile (B,f);
	bufSetFile (V,v);

	// Initialization of output buffer
	outWriter = createWriter (outfile);

	fprintf(stderr, "\nSTARTING DECOMPRESSION...[%ld]", (long)B->fpos);
	register unsigned char *pbeg,*pend,*codeStart;

	//Loading the vocabulary into memory.
	bufCurrent (V,(char**)&text,(char**)&top);
	pbeg = text; pend = top;

	{//Reads the number of words of the vocabulary (stored in the text file as a long value)
		unsigned long numbertmp;
		int z;
		unsigned char *c;
		c= (unsigned char*) &numbertmp;
		for (z=1; z<=4;z++) {
			*c =*pbeg;
			c++;
			pbeg++;
		}
		fprintf(stderr,"\n\t*Number of words in the vocabulary is: %ld",numbertmp);
		zeroNode=numbertmp;

	}

	{//Reads the value of S (and C = 256-S)
		codeStart = pbeg;
		while (*pbeg !='\0') {
			pbeg++;
		}

		{
			register unsigned long len;
			register int i;
			len = pbeg-codeStart;
			for (i=0;i<=len;i++) {  //copia tambien el '\0'
				valor_S[i]=*(codeStart++);
			}
		}
		S=atoi(valor_S);

		C=256-S;
		pbeg++;
	}

	initialize(zeroNode); //initializes decompressor for up to zeroNode words
	fprintf(stderr,"\n\t*Value of S is: %ld, C is: %ld", (long)S,(long)C);

	/*Reads one word at a time (j) and adds it to the vocabulary vector.*/
	/*Just to rebuild the vocabulary */
	j=0;
	while(1){

		while ( pbeg < pend )   {
			//reads the word
			codeStart = pbeg;
			while (*pbeg !='\0') {
				pbeg++;
			}
			//adds the word to the vocabulary
			{
				register unsigned long len;
				unsigned char *w;
				register int i;

				len = pbeg-codeStart;
				getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);

				w=words[j].word;
				for (i=0;i<=len;i++) {  //copia tambien el '\0'
					*(w++)=*(codeStart++);
				}
				words[j].size=len;
			}

			j++;
			pbeg++;
		}

		nextbuf=pbeg;

		bufLoad (V, (char*)nextbuf);
		if (bufEmpty(V)) break;
		bufCurrent (V,(char**)&text,(char**)&top);
		pbeg = text; pend = top;
	}
	//fprintf(stderr,"\n %s\n",words[75680].word);

	//Performs the decoding process over all the codewords of the compressed file
	{
		register unsigned long j;
		register int s,c;
		s=S;
		c=C;
		while (!bufEmpty(B) ) {

			bufCurrent (B,(char**)&text,(char**)&top);
			pbeg = text; pend = top;

			while ( pbeg < pend )   {
				//Reads and decodes a codeword
				codeStart = pbeg;
				j=0;
				while (*pbeg >=s) {
					j = j * c + (*pbeg - s);
					pbeg++;
				}
				j = j * s + *pbeg;
				j += TABLABASE[pbeg - codeStart];  //the position of the word in the vocabulary

				pbeg++;
				writeWord(outWriter,words[j].word, words[j].size);
			}/* inner while*/

			nextbuf=pbeg;
			bufLoad (B, (char*)nextbuf);
		}// inner while
	}

	fprintf(stderr, "\nDECOMPRESSION FINISHED...");

	//frees buffers and memory.
	bufDestroy(B);
	
	bufDestroy(V);
	destroyWriter(outWriter);
	free(words);
	free(voc);
	destroyMemoryManager(_memMgr);

} //mainDecompressor


/* Shows a message with info of this version of the code */
void showMessage() {
	printf("\n*(S,C)-Dense Code v0.9 [semistatic decompressor]");
	printf("\n*CopyRight(c)2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param\n");
	
    printf("\n  #SCDC comes with ABSOLUTELY NO WARRANTY; ");
    printf("\n  #This is free software, and you are welcome to redistribute it ");
    printf("\n  #under certain conditions\n\n");
}

/*------------------------------------------------------------------ */
/*
 *  DECOMPRESSOR PROGRAM (S,C)-DC.
 */
/*------------------------------------------------------------------ */

int main(int argc, char* argv[])
{
	char fileName[20];

	showMessage();

	if (argc < 3) {
		fprintf(stderr,"\n Wrong call to SCDC  decompressor\n");
		fprintf(stderr," Use:  %s  <in file>  <out file> \n",argv[0]);
		return 0;
	}

	StartRusageTime();
	mainDecompressor(argv[1], argv[2]);
	StopRusageTime("timesDec",argv[1]);
	fprintf(stderr,"\n");
}

